// Copyright (c) 2011, Giliam de Carpentier. All rights reserved.
// See http://www.decarpentier.nl/scape-procedural-extensions/ for more info
// 
// This file is licensed under the FreeBSD license:
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
//    1. Redistributions of source code must retain the above copyright notice, this list of
//       conditions and the following disclaimer.
// 
//    2. Redistributions in binary form must reproduce the above copyright notice, this list
//       of conditions and the following disclaimer in the documentation and/or other materials
//       provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GILIAM DE CARPENTIER 'AS IS' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GILIAM DE CARPENTIER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "ProceduralBasics.h"

float3 perlinNoiseDeriv(float2 p, float seed)
{
	// Calculate 2D integer coordinates i and fraction p.
	float2 i = floor(p);
	float2 f = p - i;
	
	// Get weights from the coordinate fraction
	float2 w = f * f * f * (f * (f * 6 - 15) + 10); // 6f^5 - 15f^4 + 10f^3
	float4 w4 = float4(1, w.x, w.y, w.x * w.y);

	// Get the derivative dw/df
	float2 dw = f * f * (f * (f * 30 - 60) + 30); // 30f^4 - 60f^3 + 30f^2

	// Get the derivative d(w*f)/df
	float2 dwp = f * f * f * (f * (f * 36 - 75) + 40); // 36f^5 - 75f^4 + 40f^3

	// Get the four randomly permutated indices from the noise lattice nearest to
	// p and offset these numbers with the seed number.
	float4 perm = tex2D(samplerPerlinPerm2D, i / 256) + seed;
	
	// Permutate the four offseted indices again and get the 2D gradient for each
	// of the four permutated coordinates-seed pairs.
	float4 g1 = tex2D(samplerPerlinGrad2D, perm.xy) * 2 - 1;
	float4 g2 = tex2D(samplerPerlinGrad2D, perm.zw) * 2 - 1;
	
	// Evaluate the four lattice gradients at p
	float a = dot(g1.xy, f);
	float b = dot(g2.xy, f + float2(-1,  0));
	float c = dot(g1.zw, f + float2( 0, -1));
	float d = dot(g2.zw, f + float2(-1, -1));
	
	// Bi-linearly blend between the gradients, using w4 as blend factors.
	float4 grads = float4(a, b - a, c - a, a - b - c + d);
	float n = dot(grads, w4);

	// Calculate the derivatives dn/dx and dn/dy
	float dx = (g1.x + (g1.z-g1.x)*w.y) + ((g2.y-g1.y)*f.y - g2.x + ((g1.y-g2.y-g1.w+g2.w)*f.y + g2.x + 
	           g1.w - g2.z - g2.w)*w.y)*dw.x + ((g2.x-g1.x) + (g1.x-g2.x-g1.z+g2.z)*w.y)*dwp.x;
	float dy = (g1.y + (g2.y-g1.y)*w.x) + ((g1.z-g1.x)*f.x - g1.w + ((g1.x-g2.x-g1.z+g2.z)*f.x + g2.x + 
	           g1.w - g2.z - g2.w)*w.x)*dw.y + ((g1.w-g1.y) + (g1.y-g2.y-g1.w+g2.w)*w.x)*dwp.y;

	return float3(n, dx, dy) * 1.5;
}

float swissTurbulence(float2 p, float seed, int octaves,
                      float lacunarity = 2.0, float gain = 0.5,
					  float warp = 0.15)
{
     float sum = 0;
     float freq = 1.0, amp = 1.0;
     float2 dsum = float2(0,0);
     for(int i=0; i < octaves; i++)
     {
         float3 n = perlinNoiseDeriv((p + warp * dsum)*freq, seed + i / 256.0);
         sum += amp * (1 - abs(n.x));
         dsum += amp * n.yz * -n.x;
         freq *= lacunarity;
         amp *= gain * saturate(sum);
    }
    return sum;
}

float jordanTurbulence(float2 p, float seed, int octaves, float lacunarity = 2.0,
                       float gain1 = 0.8, float gain = 0.5,
                       float warp0 = 0.4, float warp = 0.35,
                       float damp0 = 1.0, float damp = 0.8,
                       float damp_scale = 1.0)
{
    float3 n = perlinNoiseDeriv(p, seed);
    float3 n2 = n * n.x;
    float sum = n2.x;
    float2 dsum_warp = warp0*n2.yz;
    float2 dsum_damp = damp0*n2.yz;

    float amp = gain1;
    float freq = lacunarity;
    float damped_amp = amp * gain;

    for(int i=1; i < octaves; i++)
    {
        n = perlinNoiseDeriv(p * freq + dsum_warp.xy, seed + i / 256.0);
        n2 = n * n.x;
        sum += damped_amp * n2.x;
        dsum_warp += warp * n2.yz;
        dsum_damp += damp * n2.yz;
        freq *= lacunarity;
        amp *= gain;
        damped_amp = amp * (1-damp_scale/(1+dot(dsum_damp,dsum_damp)));
    }
    return sum;
}

float gDistortionScale
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.01;
    string UIName =  "Distortion scale";
> = 0.15f;

float gDistortionGain
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.01;
    string UIName =  "Distortion gain";
> = 0.5f;

float gDistortionUVScale
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 100.0;
    float UIStep = 0.01;
    string UIName =  "Distortion UV Scale";
> = 4.0f;

#ifdef DEFINE_PS_SWISSTURBULENCE

float gWarp
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.01;
    string UIName =  "Warp";
> = 0.15f;

float4 psSwissTurbulence(vertexOutput IN) : COLOR
{
	float2 p = IN.UV * gUVScale;
	p.x += gDistortionScale * turbulence(IN.UV * gDistortionUVScale, gSeed + 0.1, 4, gLacunarity, gDistortionGain);
	p.y += gDistortionScale * turbulence(IN.UV * gDistortionUVScale, gSeed + 0.2, 4, gLacunarity, gDistortionGain);
	float height = swissTurbulence(p, gSeed, OCTAVES, gLacunarity, gGain, gWarp); 
	return 0.25f * height - 0.05f;
}

#endif // DEFINE_PS_SWISSTURBULENCE

#ifdef DEFINE_PS_JORDANTURBULENCE

float gGain1
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 2.0;
    float UIStep = 0.01;
    string UIName =  "Gain1";
> = 0.8f;

float gWarp0
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 2.0;
    float UIStep = 0.01;
    string UIName =  "Warp0";
> = 0.4f;

float gWarp
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 2.0;
    float UIStep = 0.01;
    string UIName =  "Warp";
> = 0.35f;

float gDamp0
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.01;
    string UIName =  "Damp 0";
> = 1.0f;

float gDamp
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 10.0;
    float UIStep = 0.01;
    string UIName =  "Damp";
> = 0.8f;

float gDampScale
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 10.0;
    float UIStep = 0.01;
    string UIName =  "Damp scale";
> = 1.0f;

float4 psJordanTurbulence(vertexOutput IN) : COLOR
{
	float2 p = IN.UV * gUVScale;
	p.x += gDistortionScale * turbulence(IN.UV * gDistortionUVScale, gSeed + 0.1, 4, gLacunarity, gDistortionGain);
	p.y += gDistortionScale * turbulence(IN.UV * gDistortionUVScale, gSeed + 0.2, 4, gLacunarity, gDistortionGain);
	float height = jordanTurbulence(p, gSeed, OCTAVES, gLacunarity, gGain1, gGain, gWarp0, gWarp, gDamp0, gDamp, gDampScale);
	return height;
}

#endif // DEFINE_PS_JORDANTURBULENCE
